(ns com.aurellem.gb.hxc
  (:use (com.aurellem.gb assembly characters gb-driver util
                         constants))
  (:use (com.aurellem.world practice))
  (:import [com.aurellem.gb.gb_driver SaveState]))




; ************* HANDWRITTEN CONSTANTS

(def pkmn-types
  [:normal
   :fighting
   :flying
   :poison
   :ground
   :rock
   :bird
   :bug
   :ghost
   :A
   :B
   :C
   :D
   :E
   :F
   :G
   :H
   :I
   :J
   :K
   :fire
   :water
   :grass
   :electric
   :psychic
   :ice
   :dragon
   ])


;; question: when status effects claim to take
;; their accuracy from the move accuracy, does
;; this mean that the move always "hits" but the
;; status effect may not?

(def move-effects
  ["normal damage"
   "no damage, just opponent sleep" ;; how many turns? is atk power ignored?
   "0x4C chance of poison"
   "leech half of inflicted damage"
   "0x19 chance of burn"
   "0x19 chance of freeze"
   "0x19 chance of paralyze"
   "user faints; opponent defense halved."
   "leech half of inflicted damage ONLY if sleeping opponent."
   "imitate last attack"
   "user atk +1"
   "user def +1"
   "user spd +1"
   "user spc +1"
   "user acr +1" ;; unused?!
   "user evd +1"
   "get post-battle $ = 2*level*uses"
   "0xFE acr, no matter what."
   "opponent atk -1"     ;; acr taken from move acr?
   "opponent def -1"     ;;
   "opponent spd -1"   ;;
   "opponent spc -1" ;;
   "opponent acr -1";;
   "opponent evd -1"
   "converts user's type to opponent's."
   "(haze)"
   "(bide)"
   "(thrash)"
   "(teleport)"
   "(fury swipes)"
   "attacks 2-5 turns" ;; unused? like rollout?
   "0x19 chance of flinch"
   "opponent sleep for 1-7 turns"
   "0x66 chance of poison"
   "0x4D chance of burn"
   "0x4D chance of freeze"
   "0x4D chance of paralyze"
   "0x4D chance of flinch"
   "one-hit KO"
   "charge one turn, atk next."
   "fixed damage, leaves 1HP." ;; how is dmg determined?
   "fixed damage."             ;; cf seismic toss, dragon rage, psywave.
   "atk 2-5 turns; opponent can't attack" ;; unnormalized? (0 0x60 0x60 0x20 0x20)
   "charge one turn, atk next. (can't be hit when charging)"
   "atk hits twice."
   "user takes 1 damage if misses."
   "evade status-lowering effects" ;;caused by you or also your opponent?
   "(broken) if user is slower than opponent, makes critical hit impossible, otherwise has no effect"
   "atk causes recoil dmg = 1/4 dmg dealt"
   "confuses opponent" ;; acr taken from move acr
   "user atk +2"
   "user def +2"
   "user spd +2"
   "user spc +2"
   "user acr +2" ;; unused!
   "user evd +2" ;; unused!
   "restores up to half of user's max hp." ;; broken: fails if the difference
   ;; b/w max and current hp is one less than a multiple of 256.
   "(transform)"
   "opponent atk -2"
   "opponent def -2"
   "opponent spd -2"
   "opponent spc -2"
   "opponent acr -2"
   "opponent evd -2"
   "doubles user spc when attacked"
   "doubles user def when attacked"
   "just poisons opponent"   ;;acr taken from move acr
   "just paralyzes opponent" ;;
   "0x19 chance opponent atk -1"
   "0x19 chance opponent def -1"
   "0x19 chance opponent spd -1"
   "0x4C chance opponent spc -1" ;; context suggest chance is 0x19
   "0x19 chance opponent acr -1"
   "0x19 chance opponent evd -1"
   "???" ;; unused? no effect?
   "???" ;; unused? no effect?
   "0x19 chance opponent confused"
   "atk hits twice. 0x33 chance opponent poisioned."
   "broken. crash the game after attack."
   "(substitute)"
   "unless opponent faints, user must recharge after atk. some
  exceptions apply."
   "(rage)"
   "(mimic)"
   "(metronome)"
   "(leech seed)"
   "does nothing (splash)"
   "(disable)"
   ])


;; ************** HARDCODED DATA

(defn hxc-thunk
  "Creates a thunk (unary fn) that grabs data in a certain region of rom and
splits it into a collection by 0x50. If rom is not supplied, uses the
  original rom data."
  [start length]
  (fn self
    ([rom]
       (take-nth 2
                 (partition-by #(= % 0x50)
                               (take length
                                     (drop start rom)))))
    ([]
      (self com.aurellem.gb.gb-driver/original-rom))))

(def hxc-thunk-words
  "Same as hxc-thunk, except it interprets the rom data as characters,
  returning a collection of strings."
  (comp
   (partial comp (partial map character-codes->str))
   hxc-thunk))


;; --------------------------------------------------

(def hxc-items
  "The hardcoded names of the items in memory. List begins at
ROM@045B7"
  (hxc-thunk-words 0x45B7 870))

(def hxc-types
  "The hardcoded type names in memory. List begins at ROM@27D99,
  shortly before hxc-titles."
  (hxc-thunk-words 0x27D99 102))

(def hxc-titles
  "The hardcoded names of the trainer titles in memory. List begins at
ROM@27E77"
  (hxc-thunk-words 0x27E77 196))

(def hxc-places
  "The hardcoded place names in memory. List begins at
ROM@71500. [Cinnabar] Mansion seems to be dynamically calculated."
  (hxc-thunk-words 0x71500 560))


(defn hxc-dialog
  "The hardcoded dialogue in memory, including in-game alerts. Dialog
  seems to be separated by 0x57 instead of 0x50 (END). Begins at ROM@98000."
  ([rom]
     (map character-codes->str
          (take-nth 2
                    (partition-by #(= % 0x57)
                                  (take 0x0F728
                                        (drop 0x98000 rom))))))
  ([]
     (hxc-dialog com.aurellem.gb.gb-driver/original-rom)))


(def hxc-pokedex
  "The hardcoded pokedex entries in memory. List begins at
ROM@B8000, shortly before move names."
  (hxc-thunk-words 0xB8000 14754))

(def hxc-move-names
  "The hardcoded move names in memory. List begins at ROM@BC000"
  (hxc-thunk-words 0xBC000 1551))


(defn hxc-move-data
  "The hardcoded (basic (move effects)) in memory. List begins at
0x38000. Returns a map of {:name :power :accuracy :pp :fx-id
  :fx-txt}. The move descriptions are handwritten, not hardcoded."
  ([]
     (hxc-move-data com.aurellem.gb.gb-driver/original-rom))
  ([rom]
     (let [names (vec (hxc-move-names rom))
           move-count (count names)
           move-size 6
           format-name (fn [s]
                      (keyword (.toLowerCase
                                (apply str
                                       (map #(if (= % \space) "-" %) s)))))]
       (zipmap (map format-name names)
               (map
                (fn [[idx effect power type accuracy pp]]
                  {:name (names (dec idx))
                   :power power
                   :accuracy  accuracy
                   :pp pp
                   :fx-id effect
                   :fx-txt (get move-effects effect) 
                   }
                  )
                
                (partition move-size
                           (take (* move-size move-count)
                                 (drop 0x38000 rom))))))))
  


(defn hxc-move-data*
  "Like hxc-move-data, but reports numbers as hexadecimal symbols instead."
  ([]
     (hxc-move-data* com.aurellem.gb.gb-driver/original-rom))
  ([rom]
     (let [names (vec (hxc-move-names rom))
           move-count (count names)
           move-size 6
           format-name (fn [s]
                         (keyword (.toLowerCase
                                   (apply str
                                          (map #(if (= % \space) "-" %) s)))))
           ]
       (zipmap (map format-name names)
               (map
                (fn [[idx effect power type accuracy pp]]
                  {:name (names (dec idx))
                   :power power
                   :accuracy (hex accuracy)
                   :pp pp
                   :fx-id (hex effect)
                   :fx-txt (get move-effects effect) 
                   }
                  )
                
                (partition move-size
                           (take (* move-size move-count)
                                 (drop 0x38000 rom))))))))
  


(defn hxc-pokenames
  "The hardcoded names of the 190 species in memory. List begins at
ROM@E8000. Although names in memory are padded with 0x50 to be 10 characters
  long, these names are stripped of padding."
  ([]
     (hxc-pokenames com.aurellem.gb.gb-driver/original-rom))
  ([rom]
     (let [count-species 190
           name-length 10]
       (map character-codes->str
            (partition name-length
                       (map #(if (= 0x50 %) 0x00 %)
                            (take (* count-species name-length)
                                  (drop 0xE8000
                                        rom))))))))

(defn hxc-advantage
  "The hardcoded type advantages in memory, returned as tuples of atk-type def-type multiplier. By default (i.e. if not listed here),
the multiplier is 1."
  ([] (hxc-advantage com.aurellem.gb.gb-driver/original-rom))
  ([rom]
     (map
      (fn [[atk def mult]] [(get pkmn-types atk (hex atk))
                            (get pkmn-types def (hex def))
                            (/ mult 10)])
      (partition 3
                 (take-while (partial not= 0xFF)
                             (drop 0x3E62D rom))))))










;; ********************** MANIPULATION FNS




(defn submap?
  "Compares the two maps. Returns true if map-big has the same associations as map-small, otherwise false." 
  [map-small map-big]
  (cond (empty? map-small) true
        (and
         (contains? map-big (ffirst map-small))
         (= (get map-big (ffirst map-small))
            (second (first map-small))))
        (recur (next map-small) map-big)

        :else false))  


(defn search-map [proto-map maps]
  "Returns all the maps that make the same associations as proto-map."
  (some (partial submap? proto-map) maps))

;; (I don't use this for anything so far.)
;;
;; (defn filter-vals
;;   "Returns a map consisting of all the pairs [key val] for which (pred
;;   key) returns true."
;;   [pred map]
;;   (reduce (partial apply assoc) {} (filter (fn [[k v]] (pred v)) map)))


(defn search-moves
  "Returns a subcollection of all hardcoded moves with the given
  attributes. Attributes consist of :name :power :accuracy :pp :fx-id
  (and also :fx-txt, but it contains the same information as :fx-id)" 
  ([attribute-map]
     (search-moves com.aurellem.gb.gb-driver/original-rom attribute-map))
  ([rom attribute-map]
     (filter-vals (partial submap? attribute-map) (hxc-move-data
                                                   rom))))

  









;; note for later: credits start at F1290



(comment 

(def hxc-later
  "Running this code produces, e.g. hardcoded names NPCs give
their pokemon. Will sort through it later."
(print (character-codes->str(take 10000
                          (drop 0x71597
                                (rom (root)))))))

(let [dex
      (partition-by #(= 0x50 %)
                    (take 2540
                          (drop 0x40687
                                (rom (root)))))]
  (def dex dex)
  (def hxc-species
    (map character-codes->str
         (take-nth 4 dex))))
)